from __future__ import with_statement
import sys, traceback, time
from threading import currentThread as current_thread

__all__ = [
    'Logger', 'StreamHandler', 'CallbackHandler', 'MultiHandler',
    'NullHandler', 'NullLogger',
    'root_log', 'stdout_log',
]

level_names = {}
level_by_name = {}

def add_level(name, level):
    assert name not in level_by_name
    level_by_name[name] = level
    level_names[level] = name

for i, name in enumerate('DEBUG INFO WARNING ERROR CRITICAL'.split()):
    add_level(name, 10*(i+1))

#@@ handle encoding errors (log as repr)
#@@ handle logging errors in multihandler
#@@ handle logging errors on top level

#print level_by_name['CRITICAL'] / 10 - 1

class LogRecord(dict):
    def __getitem__(self, key):
        try:
            return dict.__getitem__(self, key)
        except KeyError:
            get = getattr(self, 'get_' + key, None)
            if get is None:
                raise
            val = self[key] = get()
            return val

    def get_level_name(self):
        return level_names[self['level']]

    def get_logger_name(self):
        return self['logger'].name

    def get_thread_name(self):
        return self['thread'].getName()

    def get_traceback(self):
        exc_info = self['exc_info']
        if exc_info is not None:
            return ''.join(traceback.format_exception(*exc_info))

    def get_full_message(self):
        #return '%s\n%s' % (self['message'], self['traceback'] or '')
        return '%s\n%s' % (self['message'], self['traceback'] or '')

    def get_timestamp_fmt(self):
        return "%Y-%m-%d %H:%M:%S"

    def get_timestamp_str(self):
        ct = time.localtime(self['timestamp'])
        return time.strftime(self['timestamp_fmt'], ct)

    def __repr__(self):
        return '%s(%s)' % (self.__class__.__name__, dict.__repr__(self))

    #def __str__(self):
    #    return "%(message)s\n%(traceback)s" % self






class LogHandlerBase(object):
    def _set_min_level(self, level):
        if not isinstance(level, int):
            level = level_by_name[level.upper()]
        self._min_level = level
    min_level = property(lambda self: self._min_level, _set_min_level)

    def handle_rec(self, rec):
        if rec['level'] >= self._min_level:
            self._handle_rec(rec)

    def _handle_rec(self, rec):
        raise NotImplementedError

class LeafLogHandlerBase(LogHandlerBase):
    format = u"%(level_name)s:%(logger_name)s:%(full_message)s"
    _min_level = 0

    def __init__(self, min_level=None, format=None):
        if min_level is not None:
            self.min_level = min_level
        if format is not None:
            self.format = unicode(format)


class StreamHandler(LeafLogHandlerBase):
    def __init__(self, stream, min_level=None, format=None, encoding='UTF-8'):
        self.stream = stream
        self.encoding = encoding
        LeafLogHandlerBase.__init__(self, min_level, format)

    def _handle_rec(self, rec):
        self.stream.write((self.format % rec).encode(self.encoding).replace('\n', '\r\n'))



class CallbackHandler(LogHandlerBase):
    def __init__(self, cb, min_level=0):
        self.cb = cb
        self.min_level = min_level

    def _handle_rec(self, rec):
        self.cb(rec)

class MultiHandler(LogHandlerBase):
    def __init__(self, items=(), min_level=0):
        self.handlers = []
        map(self.add, items)
        self.min_level = min_level

    def add(self, handler):
        if not hasattr(handler, 'handle_rec'):
            assert hasattr(handler, '__call__')
            handler = CallbackHandler(handler)
        assert handler not in self.handlers
        self.handlers.append(handler)

    def remove(self, handler):
        self.handlers.remove(handler)

    def _handle_rec(self, rec):
        for handler in self.handlers:
            handler.handle_rec(rec)


class StdoutHandler(StreamHandler):
    stream = property(lambda self: sys.stdout)
    def __init__(self, min_level=None, format=None, encoding='UTF-8'):
        self.encoding = encoding
        LeafLogHandlerBase.__init__(self, min_level, format)


stdout_log = StdoutHandler()
root_log = MultiHandler([stdout_log], 0)#@@'WARNING')



def get_caller():
    """
        -> fn, lineno, func_name
    """
    fr = sys._getframe(3)
    code = fr.f_code
    assert code.co_filename != __file__
    return code.co_filename, fr.f_lineno, code.co_name




class Logger(LogHandlerBase):
    def __init__(self, name, parent=root_log, min_level=0):
        self.name = name
        self.parent = parent
        self.min_level = min_level

    def _log(self, level, args, kw=None):
        if level < self._min_level:
            return
        elif kw:
            assert len(kw) == 1
            exc_info = kw['exc_info']
            if not isinstance(exc_info, tuple):
                assert isinstance(exc_info, bool)
                exc_info = sys.exc_info() if exc_info else None
            if exc_info == (None, None, None):
                exc_info = None
        else:
            exc_info = None
        msg = unicode(args[0])
        if len(args) > 1:
            msg %= args[1:]
        fn, line_no, func_name = get_caller()
        rec = LogRecord({
            'logger': self, 'level': level, 'message': msg, #'args': args,
            'filename': fn, 'line_no': line_no, 'func_name': func_name,

            'exc_info': exc_info,
            'timestamp': time.time(),
            'thread': current_thread(),
            #'pid': os.getpid()
        })
        self.parent.handle_rec(rec)

    def _handle_rec(self, rec):
        self.parent.handle_rec(rec)

    def debug(self, *args): self._log(10, args)
    def info(self, *args): self._log(20, args)
    def warning(self, *args): self._log(30, args)
    def error(self, *args, **kw): self._log(40, args, kw)
    def critical(self, *args, **kw): self._log(50, args, kw)

    #warn = warning
    #fatal = critical



class NullHandler(object):
    __slots__ = ()
    def handle_rec(self, rec):
        pass


class NullLogger(NullHandler):
    __slots__ = ()
    def _noop(self, *args): pass
    def _noop_kw(self, *args, **kw): pass
    debug = info = warning = _noop
    error = critical = _noop_kw
